###################################################
# Principles of Business Forecasting 2e, 2017
# ord, K., Fildes, R. and Kourentzes, N.
# (Book website) - TO BE EDITED
#
# Companion scripts for R
###################################################

# Exercises 3.13

###################################################
# These line will be removed when this is packaged!
load("./data/data_SA_oil_prices_2.Rdata")
###################################################

# To run the damped trend model we will use the ets function and set model="AAN" and damped=TRUE
# For the log-transfrom we need to preprocess the time series with the function log and
# post process the forecasts with the function exp. Always compare model performance on the
# same data!
if (!require("forecast")){install.packages("forecast")}

# Store the data in a conveniently named variable
y <- data_SA_oil_prices_2
# We can use the window function to change the length of the series
# Here we will solve it for one case, using the same approach we can solve the second estimation sample
y <- window(y,start=1978)  # Set the starting date of the series
n <- length(y)

# Let us find how long is the evaluation set
test <- window(y,start=2005) # Set the start date of the test set, this will also limit the fitting sample
m <- length(test)

# Fit models
k <- 2
fit <- vector("list",k)
# Damped trend exponential smoothing
fit[[1]] <- ets(window(y,end=2004),model="AAN",damped=TRUE)
# Linear trend exponential smoothing on log-transformed data
fit[[2]] <- ets(log(window(y,end=2004)),model="AAN",damped=FALSE)

# Initialise an array to store the forecasts
frc <- array(NA,c(m,k),dimnames=list(time(test),c("DES","LES-log")))

# Produce forecasts
for (i in 1:m){
  # We run a rolling origin evaluation. At each iteration (j) we need to provide the current fitting sample
  # This must be a ts class object for the forecast package functions to work properly
  fitsample <- y[1:(n-m+i-1)]
  fitsample <- ts(fitsample,frequency=frequency(y),start=start(y))
  for (j in 1:k){
    # First we fit ets with the fixed alpha value and model="AAN" - we will let it optimise the initial trend
    if (j == 1){
      fitrol <- ets(fitsample,model=fit[[j]],use.initial.values=TRUE)
      frc[i,j] <- forecast(fitrol,h=1)$mean
    } else {
      fitrol <- ets(log(fitsample),model=fit[[j]],use.initial.values=TRUE)
      frc[i,j] <- exp(forecast(fitrol,h=1)$mean)
    }
  }
}

# To simplify plotting the results we convert frc into a ts object
frc <- ts(frc,frequency=frequency(test),start=start(test))

# Produce plot
if (!require("RColorBrewer")){install.packages("RColorBrewer")}
cmp <- brewer.pal(max(3,k),"Set1")

plot(y,ylab="Oil prices")
for (j in 1:k){
  lines(frc[,j],col=cmp[j])
}
legend("topleft",c("Data",colnames(frc)),col=c("black",cmp),lty=1)

# Calculate error metrics
e <- matrix(rep(test,k),ncol=k) - frc
# Error metrics
RMSE <- sqrt(apply(e^2,2,mean))
RMdSE <- sqrt(apply(e^2,2,median))
MAE <- apply(abs(e),2,mean)
MdAE <- apply(abs(e),2,median)
MAPE <- apply(abs(e)/matrix(rep(test,k),ncol=k),2,mean)*100
MdAPE <- apply(abs(e)/matrix(rep(test,k),ncol=k),2,median)*100
E <- rbind(RMSE,MAE,MAPE,RMdSE,MdAE,MdAPE)
print(round(E,3))

# We will provide PIs for the LES-log model. For DES you can repeat from exercise 3.9
# We construct the PIs on the log-transformed data and then revert them back.
# This will construct assymetric PIs
e <- fit[[2]]$residuals
RMSE <- sqrt(mean(e^2))
a <- 0.95
PIlower <- exp(log(frc[,k]) - abs(qnorm((1-a)/2))*RMSE)
PIupper <- exp(log(frc[,k]) + abs(qnorm((1-a)/2))*RMSE)

# Compare with actuals
upper <- test > PIupper
lower <- test < PIlower
# Dates that actuals are outside PIs
time(test)[which(upper | lower)]

# Visualise the comparison
yy <- range(c(test,PIupper,PIlower))
plot(test,ylim=yy,ylab="Oil prices")
lines(frc[,2],col=cmp[2])
lines(PIupper,col=cmp[2],lty=2)
lines(PIlower,col=cmp[2],lty=2)
points(time(test)[which(upper|lower)],test[which(upper|lower)],pch=19,col=cmp[1])
legend("topleft",c("Out-of-sample","Forecast","Prediction Intervals","Outside PIs"),col=c("black",cmp[2],cmp[2],cmp[1]),lty=c(1,1,2,1),pch=c(NA,NA,NA,19))
