###################################################
# Principles of Business Forecasting 2e, 2017
# ord, K., Fildes, R. and Kourentzes, N.
# (Book website) - TO BE EDITED
#
# Companion scripts for R
###################################################

# Exercises 3.9

###################################################
# These line will be removed when this is packaged!
load("./data/data_SA_oil_prices_2.Rdata")
###################################################

# Store the data in a conveniently named variable
y <- data_SA_oil_prices_2
# We can use the window function to change the length of the series
# Here we will solve it for one case, using the same approach we can solve the second estimation sample
y <- window(y,start=1978)  # Set the starting date of the series
n <- length(y)

# Let us find how long is the evaluation set
test <- window(y,start=2005) # Set the start date of the test set, this will also limit the fitting sample
m <- length(test)

# we produce linear exponential smoothing forecasts using the ets function
# We need to set model="AAN"
# Since we are not provided with instructions about the value of beta, we will optimise it
if (!require("forecast")){install.packages("forecast")}

# Find optimal alpha
alpha_opt <- ets(window(y,end=2007),model="AAN")$par[1]
alpha <- c(0.2,0.5,0.8,alpha_opt)
k <- length(alpha)

# Find optimal betas. Note that optimal alpha and beta could be found in a single step.
beta <- vector("numeric",k)
for (i in 1:k){
  fit <- ets(window(y,end=time(test)[1]-1),model="AAN",alpha=alpha[i])
  pars <- fit$par
  beta[i] <- pars[which(names(pars)=="beta")]
}
# I will save the last model fit (optimal values) for extracting information about the prediction intervals
fit_opt <- fit

# Initialise an array to store the forecasts
frc <- array(NA,c(m,k),dimnames=list(time(test),paste0("AAN (a=",trunc(alpha*100)/100,",b=",trunc(beta*100)/100,")")))

# Produce forecasts
for (i in 1:m){
  # We run a rolling origin evaluation. At each iteration (j) we need to provide the current fitting sample
  # This must be a ts class object for the forecast package functions to work properly
  fitsample <- y[1:(n-m+i-1)]
  fitsample <- ts(fitsample,frequency=frequency(y),start=start(y))
  for (j in 1:k){
    # First we fit ets with the fixed alpha value and model="AAN" - we will let it optimise the initial trend
    fit <- ets(fitsample,model="AAN",alpha=alpha[j],beta=beta[j])
    # Now we change the initial value in the fitted model...
    fit$initstate[1] <- window(y,start=time(y)[1],end=time(y)[1])
    # ... and recalculate
    fit <- ets(fitsample,model=fit,use.initial.values=TRUE)
    # This trick can work with any exponential smoothing model

    # Produce the 1-step ahead forecast
    frc[i,j] <- forecast(fit,h=1)$mean
  }
}

# To simplify plotting the results we convert frc into a ts object
frc <- ts(frc,frequency=frequency(test),start=start(test))

# Produce plot
if (!require("RColorBrewer")){install.packages("RColorBrewer")}
cmp <- brewer.pal(k,"Set1")

plot(y,ylab="Oil prices")
for (j in 1:k){
  lines(frc[,j],col=cmp[j])
}
legend("topleft",c("Data",colnames(frc)),col=c("black",cmp),lty=1)

# Calculate error metrics
e <- matrix(rep(test,k),ncol=k) - frc
# Error metrics
RMSE <- sqrt(apply(e^2,2,mean))
RMdSE <- sqrt(apply(e^2,2,median))
MAE <- apply(abs(e),2,mean)
MdAE <- apply(abs(e),2,median)
MAPE <- apply(abs(e)/matrix(rep(test,k),ncol=k),2,mean)*100
MdAPE <- apply(abs(e)/matrix(rep(test,k),ncol=k),2,median)*100
E <- rbind(RMSE,MAE,MAPE,RMdSE,MdAE,MdAPE)
print(round(E,3))

# To calculate the prediction intervals we need the RMSE of the model fit.
# We can easily get this from the ets model fit. However, as we will see in
# Chapter 4, ets in fact provides model based prediction intervals. We will
# not use these now.

# We will provide PIs for the optimal fit model. The rest can be done in the same way
e <- fit_opt$residuals
RMSE <- sqrt(mean(e^2))
a <- 0.95
PIlower <- frc[,k] -abs(qnorm((1-a)/2))*RMSE
PIupper <- frc[,k] +abs(qnorm((1-a)/2))*RMSE

# Compare with actuals
upper <- test > PIupper
lower <- test < PIlower
# Dates that actuals are outside PIs
time(test)[which(upper | lower)]

# Visualise the comparison
yy <- range(c(test,PIupper,PIlower))
plot(test,ylim=yy,ylab="Oil prices")
lines(frc[,4],col=cmp[2])
lines(PIupper,col=cmp[2],lty=2)
lines(PIlower,col=cmp[2],lty=2)
points(time(test)[which(upper|lower)],test[which(upper|lower)],pch=19,col=cmp[1])
legend("topleft",c("Out-of-sample","Forecast","Prediction Intervals","Outside PIs"),col=c("black",cmp[2],cmp[2],cmp[1]),lty=c(1,1,2,1),pch=c(NA,NA,NA,19))
